import threading
import queue
import logging
import time
import random
import paramiko
from typing import List
from nornir.core.task import AggregatedResult, Task
from nornir.core.inventory import Host


log = logging.getLogger(__name__)
LOCK = threading.Lock()


class RetryRunner:
    """
    RetryRunner is a modification of QueueRunner that strives to
    make task execution as reliable as possible.

    Arguments:
        num_workers: number of threads for tasks execution
        num_connectors: number of threads for device connections
        connect_retry: number of connection attempts
        connect_backoff: exponential backoff timer in milliseconds
        connect_splay: random splay interval for each connection
        task_retry: number of attemptsto run task
        task_backoff: exponential backoff timer in milliseconds
        task_splay: random splay interval before task start
    """

    def __init__(
        self,
        num_workers: int = 100,
        num_connectors: int = 20,
        connect_retry: int = 3,
        connect_backoff: int = 5000,
        connect_splay: int = 100,
        task_retry: int = 1,
        task_backoff: int = 5000,
        task_splay: int = 100
    ) -> None:
        self.num_workers = num_workers
        self.num_connectors = num_connectors
        self.connectors_q = None
        self.work_q = None
        self.connect_retry = connect_retry
        self.connect_backoff = connect_backoff
        self.connect_splay = connect_splay
        self.task_retry = task_retry
        self.task_backoff = task_backoff
        self.task_splay = task_splay
        self.jumphosts_connections = {}

    def _connect_to_device_behind_jumphost(self, host):
        """
        Establish connection to devices behind jumphost/bastion
        """

        jumphost = {"timeout": 3, "look_for_keys": False, "allow_agent": False}
        jumphost.update(host["jumphost"])
        # Initiate connection to jumphost if not initiated already or
        # if it failed before
        #
        # self.jumphosts_connections dictionary shared between threads,
        # first thread to start will detect that no connection to
        # jumphost exists and will try to establish it, while doing
        # so, jumphost key will be added to self.jumphosts_connections
        # dictionary with '__connecting__' value, next threads
        # will see '__connecting__' as value for jumphost and will
        # continue sleeping until either connection succeeded or
        # failed. On connection failure, first thread will set value
        # to '__failed__' to signal other threads to exit.
        if (
            jumphost["hostname"] not in self.jumphosts_connections or
            self.jumphosts_connections.get(jumphost["hostname"]) == "__failed__"
        ):
            try:
                LOCK.acquire()
                self.jumphosts_connections[jumphost["hostname"]] = "__connecting__"
                LOCK.release()
                jumphost_ssh_client = paramiko.client.SSHClient()
                jumphost_ssh_client.set_missing_host_key_policy(
                    paramiko.AutoAddPolicy()
                )
                jumphost_ssh_client.connect(**jumphost)
                LOCK.acquire()
                self.jumphosts_connections[jumphost["hostname"]] = {
                    "jumphost_ssh_client": jumphost_ssh_client,
                    "jumphost_ssh_transport": jumphost_ssh_client.get_transport(),
                }
                LOCK.release()
                # add jumphost to host connections to close it
                # on nornir.close_connections() call
                host.connections[
                    "jumphost_{}".format(jumphost["hostname"])
                ] = jumphost_ssh_client
                log.info(
                    "Started connection to jumphost '{}' - '{}'".format(
                        jumphost["hostname"], jumphost_ssh_client
                    )
                )
            except Exception as e:
                LOCK.acquire()
                self.jumphosts_connections[jumphost["hostname"]] = "__failed__"
                LOCK.release()
                # add exception info to host data to include in results
                error_msg = "Failed connection to jumphost '{}', error - {}".format(
                    host["jumphost"]["hostname"], e
                )
                host["exception"] = error_msg
                log.error(error_msg)
                return
        else:
            # sleep random time waiting for connection to jumphost to establish
            while self.jumphosts_connections[jumphost["hostname"]] == "__connecting__":
                time.sleep(random.randrange(0, 500) / 1000)
            if self.jumphosts_connections[jumphost["hostname"]] == "__failed__":
                # add exception info to host data to include in results
                error_msg = (
                    "Failed connection to jumphost '{}' in another thread".format(
                        host["jumphost"]["hostname"]
                    )
                )
                host["exception"] = error_msg
                log.error(error_msg)
                return
        # connect to host
        channel_name = "jumphost_{}_channel".format(jumphost["hostname"])
        if not host.connections.get(channel_name):
            dest_addr = (host.hostname, host.port or 22)
            channel = self.jumphosts_connections[jumphost["hostname"]][
                "jumphost_ssh_transport"
            ].open_channel(
                kind="direct-tcpip",
                dest_addr=dest_addr,
                src_addr=("localhost", 7777),
                timeout=3,
            )
            host.connections[channel_name] = channel
            log.info(
                "{} - started new channel via jumphost '{}' - '{}'".format(
                    host.name,
                    jumphost["hostname"],
                    self.jumphosts_connections[jumphost["hostname"]][
                        "jumphost_ssh_client"
                    ],
                )
            )
        return host.connections[channel_name]

    def connector(self):
        while True:
            connection = self.connectors_q.get(timeout=60)
            if connection is None:
                self.connectors_q.task_done()
                break
            task, host, params, result = connection
            # check if need backoff connection retry for this host
            if params.get("connection_retry", 0) > 0:
                elapsed = time.time() - params["timestamp"]
                should_wait = (params["connection_retry"] * self.connect_backoff) / 1000
                if elapsed < should_wait:
                    self.connectors_q.put(connection)
                    self.connectors_q.task_done()
                    continue
            # initiate connection to host
            connection_name = task.task.__globals__.get("CONNECTION_NAME", None)
            connection_name = task.params.pop("connection_name", connection_name)
            # on connect retry get connection name from params
            connection_name = params.get("connection_name", connection_name)
            if connection_name and connection_name not in host.connections:
                try:
                    time.sleep(random.randrange(0, self.connect_splay) / 1000)
                    if host.get("jumphost") and connection_name == "netmiko":
                        extras = {}
                        if host.connection_options.get("netmiko"):
                            extras = host.connection_options["netmiko"].extras
                        extras["sock"] = self._connect_to_device_behind_jumphost(host)
                        host.open_connection(
                            connection_name, configuration=task.nornir.config, extras=extras
                        )
                    else:
                        host.open_connection(
                            connection_name, configuration=task.nornir.config
                        )
                except Exception as e:
                    params.setdefault("connection_retry", 0)
                    params["connection_name"] = connection_name
                    log.error(
                        "{} - connection retry attempt {}, error: '{}'".format(
                            host.name, params["connection_retry"], e
                        )
                    )
                    if params["connection_retry"] < self.connect_retry:
                        params["connection_retry"] += 1
                        params["timestamp"] = time.time()
                        self.connectors_q.put(connection)
                        self.connectors_q.task_done()
                        continue
                log.info(
                    "{} - started connection: '{}'".format(host.name, connection_name)
                )
            self.connectors_q.task_done()
            self.work_q.put((task, host, params, result))

    def worker(self):
        while True:
            work = self.work_q.get(timeout=300)
            if work is None:
                self.work_q.task_done()
                break
            task, host, params, result = work
            # check if need backoff task retry for this host
            if params.get("task_retry", 0) > 0:
                elapsed = time.time() - params["timestamp"]
                should_wait = (params["task_retry"] * self.task_backoff) / 1000
                if elapsed < should_wait:
                    self.work_q.put(work)
                    self.work_q.task_done()
                    continue
            log.info("{} - running task '{}'".format(host.name, task.name))
            time.sleep(random.randrange(0, self.task_splay) / 1000)
            work_result = task.start(host)
            if work_result[0].failed:
                params.setdefault("task_retry", 0)
                log.error(
                    "{} - task execution retry attempt {} failed".format(
                        host.name, params["task_retry"]
                    )
                )
                if params["task_retry"] < self.task_retry:
                    params["task_retry"] += 1
                    params["timestamp"] = time.time()
                    self.work_q.put(work)
                    self.work_q.task_done()
                    continue
            LOCK.acquire()
            result[host.name] = work_result
            LOCK.release()
            self.work_q.task_done()
            log.info("{} - task '{}' completed".format(host.name, task.name))

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        self.connectors_q = queue.Queue()
        self.work_q = queue.Queue()
        result = AggregatedResult(task.name)
        # enqueue hosts in connectors queue
        for host in hosts:
            self.connectors_q.put((task.copy(), host, {}, result))
        # start connectors threads
        connector_threads = []
        for i in range(self.num_connectors):
            t = threading.Thread(target=self.connector, args=(), daemon=True)
            t.start()
            connector_threads.append(t)
        # start worker threads
        worker_threads = []
        for i in range(self.num_workers):
            t = threading.Thread(target=self.worker, args=(), daemon=True)
            t.start()
            worker_threads.append(t)
        # block until all work completed
        self.connectors_q.join()
        self.work_q.join()
        # stop connectors
        for i in range(self.num_connectors):
            self.connectors_q.put(None)
        for t in connector_threads:
            t.join()
        # stop workers
        for i in range(self.num_workers):
            self.work_q.put(None)
        for t in worker_threads:
            t.join()
        return result



import logging
from fnmatch import fnmatchcase
from nornir.core.filter import F


log = logging.getLogger(__name__)


def FFun(nr, **kwargs):
    """
    Inventory filters dispatcher function
    """
    ret = nr
    has_filter = False
    # check if kwargs argument given, usually
    # supplied by SALT nornir-proxy
    if kwargs.get("kwargs"):
        kwargs = kwargs["kwargs"]
    if kwargs.get("FO"):
        ret = _filter_FO(ret, kwargs.pop("FO"))
        has_filter = True
    if kwargs.get("FB"):
        ret = _filter_FB(ret, kwargs.pop("FB"))
        has_filter = True
    if kwargs.get("FG"):
        ret = _filter_FG(ret, kwargs.pop("FG"))
        has_filter = True
    if kwargs.get("FP"):
        ret = _filter_FP(ret, kwargs.pop("FP"))
        has_filter = True
    if kwargs.get("FL"):
        ret = _filter_FL(ret, kwargs.pop("FL"))
        has_filter = True
    ret.has_filter = has_filter
    return ret
    
def _filter_FO(nr, filter_data):
    """
    Function to filter hosts using Filter Object
    """
    ret = nr
    if isinstance(filter_data, dict):
        ret = ret.filter(F(**filter_data))
    elif isinstance(filter_data, list):
        ret = ret.filter(F(**filter_data[0]))
        for item in filter_data[1:]:
            filtered_hosts = nr.filter(F(**item))
            ret.inventory.hosts.update(filtered_hosts.inventory.hosts)
    return ret


def _filter_FB(ret, pattern):
    """
    Function to filter hosts by name using glob patterns
    """
    return ret.filter(filter_func=lambda h: fnmatchcase(h.name, pattern))


def _filter_FG(ret, group):
    """
    Function to filter hosts using Groups
    """
    return ret.filter(filter_func=lambda h: h.has_parent_group(group))


def _filter_FP(ret, pfx):
    """
    Function to filter hosts based on IP Prefixes
    """
    import ipaddress
    import socket
    
    socket.setdefaulttimeout(1)

    def _filter_net(host):
        # convert host's ip to ip address object
        try:
            ip_addr = ipaddress.ip_address(host.hostname)
        except ValueError:
            # try to resolve hostname using DNS
            try:
                ip_str = socket.gethostbyname(host.hostname)
                ip_addr = ipaddress.ip_address(ip_str)
            except Exception as e:
                log.error(
                    "FP failed to convert host IP '{}', error '{}'".format(
                        host.name, e
                    )
                )
                return False
        # run filtering
        for net in networks:
            if ip_addr in net:
                return True
        return False

    # make a list of network objects
    prefixes = (
        [i.strip() for i in pfx.split(",")] if isinstance(pfx, str) else pfx
    )
    networks = []
    for prefix in prefixes:
        try:
            networks.append(ipaddress.ip_network(prefix))
        except Exception as e:
            log.error(
                "FP failed to convert prefix '{}', error '{}'".format(
                    prefix, e
                )
            )
    # filter hosts
    return ret.filter(filter_func=_filter_net)


def _filter_FL(ret, names_list):
    """
    Function to filter hosts names based on list of names
    """
    names_list = (
        [i.strip() for i in names_list.split(",")]
        if isinstance(names_list, str)
        else names_list
    )
    return ret.filter(filter_func=lambda h: h.name in names_list)



from .plugins.functions import ResultSerializer
from .plugins.inventory import DictInventory
from .plugins.runners import QueueRunner
from .plugins.runners import RetryRunner

__all__ = (
    "ResultSerializer", 
    "DictInventory", 
    "QueueRunner", 
    "RetryRunner"
)


import threading
import queue
from typing import List
from nornir.core.task import AggregatedResult, Task
from nornir.core.inventory import Host


LOCK = threading.Lock()


class QueueRunner:
    """
    QueueRunner run tasks over each host using queue together
    with workers threads consuming work from work queue.

    Instead of firing up num_workes threads for each batch of hosts,
    QueueRunner starts num_workes threads once and uses queue to
    submit tasks and obtain results.

    Arguments:
        num_workers: number of threads to use
    """

    def __init__(self, num_workers: int = 20) -> None:
        self.num_workers = num_workers
        self.work_q = None

    def worker(self):
        while True:
            work_to_do = self.work_q.get()
            if work_to_do is None:
                break
            task, host, result = work_to_do
            work_result = task.copy().start(host)
            LOCK.acquire()
            result[host.name] = work_result
            LOCK.release()
            self.work_q.task_done()

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        self.work_q = queue.Queue()
        result = AggregatedResult(task.name)
        # enqueue hosts in work queue
        for host in hosts:
            self.work_q.put((task, host, result))
        # start threads
        threads = []
        for i in range(self.num_workers):
            t = threading.Thread(target=self.worker, args=(), daemon=True)
            t.start()
            threads.append(t)
        # block until all tasks are done
        self.work_q.join()
        # stop workers:
        for i in range(self.num_workers):
            self.work_q.put(None)
        for t in threads:
            t.join()
        return result



def ResultSerializer(nr_results, add_details=False):
    """
    Helper function to transform Nornir results in dictionary

    :parap add_details: boolean to indicate if results should contain more info
    """
    ret = {}
    for hostname, results in nr_results.items():
        ret[hostname] = {}
        for i in results:
            # skip task groups such as _task_foo_bar
            if i.name.startswith("_"):
                continue
            # handle errors info passed from within tasks
            elif i.host.get("exception"):
                ret[hostname][i.name] = {"exception": i.host["exception"]}
            # add results details if requested to do so
            elif add_details:
                ret[hostname][i.name] = {
                    "diff": i.diff,
                    "changed": i.changed,
                    "result": i.result,
                    "failed": i.failed,
                    "exception": str(i.exception),
                }
            # form results for the rest of tasks
            else:
                ret[hostname][i.name] = (
                    {"exception": i.exception} if i.failed else i.result
                )
    return ret